/*#======================================================================
# This source code is public domain, and may be freely distributed,
# modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
# WARRANTY OF ANY KIND.
#======================================================================
# _DY_CONTRIB_PGM_:tfmtalk.c:Transform progob talk programs
#======================================================================
# LICENSE MODULES: toolkits:
#                  programs:
#                  servers:
# AUTHORS: 
# DESCRIPTION: 
#  Purpose:
#       The program 'tfmtalk' implements molecule (i.e. SMILES string)
#       transformation via rules. The program object can be tested with the
#       version of the pipetalker also in this directory.
#======================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>

#include "dt_smiles.h"
#include "dt_smarts.h"
#include "du_pipeio.h"

#define FALSE 0
#define TRUE  1

int debug = FALSE;
/* int debug = TRUE; */

int sub(char *pat, char *new, char *src)
/*
 * Scans the string src for the occurence of the character
 * sequence pat and replaces the first one with the string
 * new[].
 *
 * It does an in-place substitution and returns TRUE if
 * a substitution was performed. The caller is responsible
 * for making the buffer src[] big enough.
 */
{
   char *buffer;
   char *match;

   buffer = calloc(strlen(src)+1, 1);
   strcpy(buffer,src);
   match = strstr(src, pat);
   if (match)
   {
      strcpy(match, new);
      strcat(src, buffer+(match-src)+strlen(pat));
      free(buffer);
      return (TRUE);
   }
   else
   {
      free(buffer);
      return (FALSE);
   }
}

dt_Handle my_sequence(dt_Handle object, int typ)
/*
 *  my_sequence() -- like dt_stream(), but returns permanent sequence
 */
{
   dt_Handle seq, obs, ob;
   seq = dt_alloc_seq();
   obs = dt_stream(object, typ);
   while (NULL_OB != (ob = dt_next(obs)))
      dt_append(seq, ob);
   dt_dealloc(obs);
   return seq;
}

void WakeUpStereo(dt_Handle mol)
/*
 * Used to trigger stereochemistry perception for molecule mol. Fixes a bug
 * in dt_arbsmiles().
 */
{
   dt_Integer len;

   dt_cansmiles(&len, mol, TRUE); /* wake up lazy evaluation of stereo */
}

dt_Handle StrToSmiles(char *str)
/*
 * Converts the string pointed to by str into a molecule object and returns it.
 */
{
   dt_Handle mol;

   mol = dt_smilin(strlen(str), str);
   WakeUpStereo(mol);

   return (mol);
}

char *DebugString(dt_Handle obj)
/*
 * Returns a pointer to the string value of obj.
 */
{
   static char buffer[10][1024];
   static int  ibuf = 0;
   dt_Integer  len;
   dt_String   s;

   ibuf = (ibuf+1) % 10;

   if (obj == NULL_OB)
      strcpy(buffer[ibuf], "NULL_OB");
   else
   {
      if (dt_type(obj) == TYP_MOLECULE)
      {
	 s = dt_arbsmiles(&len, obj, TRUE);
      }
      else
	 s = dt_stringvalue(&len, obj);
      if (s)
      {
	 strncpy(buffer[ibuf], s, len);
	 buffer[ibuf][len] = '\0';
      }
      else
	 strcpy(buffer[ibuf], "<<object has no string value>>");
   }

   return (buffer[ibuf]);
}

int CheckSmiles(dt_Handle mol, dt_Handle goodseq, dt_Handle badseq)
/*
 * Checks if the molecule mol contains at least one of the SMARTS patterns in
 * goodseq and none of the patterns in badseq.
 * It returns TRUE if the molecule passes the test and FALSE otherwise.
 */
{
   dt_Handle pat;
   dt_Handle paths;
   dt_Handle sosout;
   int found, npat;

   found = dt_count(goodseq, TYP_ANY) == 0;
   if (goodseq != NULL_OB)
   {
      dt_reset(goodseq);
      for (pat = dt_next(goodseq); pat != NULL_OB; pat = dt_next(goodseq))
      {
	 paths = dt_match(pat, mol, TRUE);
	 if (dt_count(paths, TYP_PATH) > 0)
	 {
	    found = TRUE;
	    break;
	 }
	 dt_dealloc(paths);
      }
      if (!found)
      {
	 if (debug)
	    fprintf(stderr, "CheckSmarts: molecule did not match goodlist\n");
	 return (FALSE);
      }
   }

   found = FALSE;
   if (badseq != NULL_OB)
   {
      dt_reset(badseq);
      for (pat = dt_next(badseq); pat != NULL_OB; pat = dt_next(badseq))
      {
	 paths = dt_match(pat, mol, TRUE);
	 if (dt_count(paths, TYP_PATH) > 0)
	 {
	    found = TRUE;
	    break;
	 }
	 dt_dealloc(paths);
      }
      if (found)
      {
	 if (debug) fprintf(stderr, "CheckSmarts: molecule matched badlist\n");
	 return (FALSE);
      }
   }

   if (debug) fprintf(stderr, "CheckSmarts: molecule passed SMARTS check\n");
   return (TRUE);
}

#define MAXPATH 100

int SymmetricalMatch(dt_Handle mol, dt_Handle path, dt_Handle tmp_path)
/*
 * Checks if the two paths path and tmp_path match to symmetrical sites
 * in mol. This test uses the symclass information from molecule mol to check
 * if the underlying atom (and bond) sets are symmetrical to each other.
 *
 * Currently, only the atom sets are tested. This is a reasonable appriximation
 * of the complete test.
 */
{
   int path_classes[MAXPATH];	/* symmetry class list of atoms in path */
   int class_counts[MAXPATH];
   int i, n, result;
   dt_Handle at, atseq;

   if (debug)
      fprintf(stderr,
	      "SymmetricalMatch: mol = %d, path = %d, tmp_path = %d\n",
	      mol, path, tmp_path);

   if (debug)	/* collect and print tmp_classes */
   {
      /* collect symmetry classes of atoms in path */
      atseq = my_sequence(tmp_path, TYP_ATOM);
      dt_reset(atseq); n = 0;
      for (at = dt_next(atseq); at != NULL_OB; at = dt_next(atseq))
      {
	 for (i=0; i<n; i++)
	    if (dt_symclass(at, TRUE) == path_classes[i]) break;
	 if (i<n)	/* old class => increment count */
	 {
	    class_counts[i]++;
	 }
	 else	/* new class => initialize entry */
	 {
	    n++;
	    path_classes[i] = dt_symclass(at, TRUE);
	    class_counts[i] = 1;
	 }
      }
      dt_dealloc(atseq);
      fprintf(stderr,"SymmetricalMatch: tmp_classes = ");
      for (i=0; i<n; i++) fprintf(stderr, "%d(%d) ", path_classes[i], class_counts[i]);
      fprintf(stderr,"\n");
      n = 0;
   }

   /* collect symmetry classes of atoms in path */
   atseq = my_sequence(path, TYP_ATOM);
   dt_reset(atseq); n = 0;
   for (at = dt_next(atseq); at != NULL_OB; at = dt_next(atseq))
   {
      for (i=0; i<n; i++)
	 if (dt_symclass(at, TRUE) == path_classes[i]) break;
      if (i<n)	/* old class => increment count */
      {
	 class_counts[i]++;
      }
      else	/* new class => initialize entry */
      {
	 n++;
	 path_classes[i] = dt_symclass(at, TRUE);
	 class_counts[i] = 1;
      }
   }
   dt_dealloc(atseq);

   if (debug)
   {
      fprintf(stderr,"SymmetricalMatch: classes = ");
      for (i=0; i<n; i++) fprintf(stderr, "%d(%d) ", path_classes[i], class_counts[i]);
      fprintf(stderr,"\n");
   }

   result = TRUE;
   /* walk through atoms of tmp_path and remove corresponding counts */
   atseq = my_sequence(tmp_path, TYP_ATOM);
   dt_reset(atseq);
   for (at = dt_next(atseq); at != NULL_OB; at = dt_next(atseq))
   {
      for (i=0; i<n; i++)
         if (dt_symclass(at, TRUE) == path_classes[i]) break;
      if (i<n)  /* class found => decrement count */
      {
         class_counts[i]--;
	 if (class_counts[i] < 0)	/* too many of this class */
	 {
	    result = FALSE;
	    break;
	 }
      }
      else      /* new class => report non-symmetrical match */
      {
	 result = FALSE;
	 break;
      }
   }
   dt_dealloc(atseq);

   if (debug) fprintf(stderr, "SymmetricalMatch: result = %d\n", result);

   return (result);
}

#define MAXMATCH		5
#define MAXTFM			20

#define DROP_PIECES		1
#define EXHAUSTIVE		2
#define REPLACE_TFMS		4
#define UNIQUE_MATCH		8

dt_Handle Transform(dt_Handle inmol,
		    dt_Handle trnpat, dt_Handle trnmol, dt_Integer nkeep,
		    int flags)
/*
 * Translates the molecule mol using the transformation pattern
 *      (trnpat|trnmol|nkeep)
 * It either returns the transformed molecule or the NULL_OB if an error
 * was encountered or trnpat did not match.
 */
{
   dt_Handle mol, tmpmol;
   dt_Handle pathset;
   dt_Handle paths, patoms, pbonds, atoms, matoms, mbonds, batoms;
   dt_Handle path, tmp_path;
   dt_Handle atom, matom, at1, at2;
   dt_Handle bond, mbond;
   dt_Handle *mapping;	/* Pointer to array mapping result SMILES atoms to the */
			/* atoms in trnmol */
   char *tmpsmi;
   dt_Integer tmpsmilen;
   char buffer[10000], *cp;
   int unique;

   int oldnum;
   int nmatch;
   int order;

   dt_Handle result;	/* sequence of strings to be returned */

   int i;

   if (inmol == NULL_OB  ||  trnpat == NULL_OB  ||  trnmol == NULL_OB)
      return (NULL_OB);

   mol = dt_copy(inmol);	/* work on copy of input molecule */

   nmatch = 0;
   while ((nmatch < MAXMATCH  &&  (flags & EXHAUSTIVE))   	||
	  nmatch < 1)
   {

      /* find matches */
      pathset = dt_umatch(trnpat, mol, FALSE);

      if (debug) fprintf(stderr,"%d matches found\n", dt_count(pathset, TYP_PATH));

      paths = dt_stream(pathset, TYP_PATH);

      path = dt_next(paths);
      if (path == NULL_OB)	/* no match found */
      {
	 dt_dealloc(paths);
	 break;
      }

      /* work through all matches to see if resulting matches are all symmetrical */
      unique = TRUE;
      for (tmp_path = dt_next(paths); tmp_path != NULL_OB; tmp_path = dt_next(paths))
	 if (!SymmetricalMatch(mol, path, tmp_path))
	 {
	    unique = FALSE;
	    if (debug) fprintf(stderr,"unsymmetrical match found\n");
	    break;
	 }
	 else
	 {
	    if (debug) fprintf(stderr,"symmetrical match found\n");
	 }

      if (flags & UNIQUE_MATCH && !unique)
      {
	 dt_dealloc(paths);
	 break;
      }

      dt_reset(paths); path = dt_next(paths); if (path == NULL_OB) break;

      nmatch++;

      /* create sequences of atoms matching the path */
      patoms = my_sequence(path, TYP_ATOM);
      pbonds = my_sequence(path, TYP_BOND);
      dt_dealloc(paths);

      dt_mod_on(mol);
      /* remove matched bonds */
      dt_reset(pbonds);
      bond = dt_next(pbonds);
      while (bond != NULL_OB)
      {
	 batoms = dt_stream(bond, TYP_ATOM);
	 at1 = dt_next(batoms); at2 = dt_next(batoms); dt_dealloc(batoms);
	 order = dt_bondorder(bond);
	 dt_dealloc(bond);
	 dt_setimp_hcount(at1, dt_imp_hcount(at1)+order);	/* fix hydrogen counts */
	 dt_setimp_hcount(at2, dt_imp_hcount(at2)+order);
	 bond = dt_next(pbonds);
      }

      /* remove additional atoms */
      dt_reset(patoms);
      for (i=0, atom = dt_next(patoms);
	   i < nkeep  &&  atom != NULL_OB;
	   i++, atom = dt_next(patoms))
	 ;
      while (atom != NULL_OB)
      {
	 dt_dealloc(atom);
	 atom = dt_next(patoms);
      }

      /* create atom mapping for pattern atoms staying in resulting SMILES */
      mapping = (dt_Handle *)calloc(1+dt_count(trnmol, TYP_ATOM), sizeof(dt_Handle));
      mapping[0] = NULL_OB;
      matoms = dt_origsmi_stream(trnmol); dt_reset(matoms);
      dt_reset(patoms);
      for (i=0, atom = dt_next(patoms), matom = dt_next(matoms);
	   i < nkeep  &&  atom != NULL_OB  &&  matom != NULL_OB;
	   i++, atom = dt_next(patoms), matom = dt_next(matoms))
      {
	 dt_setarborder(matom, i+1);
	 mapping[i+1] = atom;
      }

      /* don't need match paths anymore */
      dt_dealloc(patoms); dt_dealloc(pbonds);

      /* add additional atoms from trnmol */
      for (i=i, matom=matom;	/* continue after keep atoms */
	   matom != NULL_OB;
	   i++, matom = dt_next(matoms))
      {
	 dt_setarborder(matom, i+1);
	 atom = dt_addatom(mol, dt_number(matom), dt_hcount(matom));
	 mapping[i+1] = atom;
      }

      /* add bonds to transformed SMILES */
      mbonds = dt_stream(trnmol, TYP_BOND); dt_reset(mbonds);
      for (mbond = dt_next(mbonds); mbond != NULL_OB; mbond = dt_next(mbonds))
      {
	 batoms = dt_stream(mbond, TYP_ATOM);
	 at1 = mapping[dt_arborder(dt_next(batoms))];
	 at2 = mapping[dt_arborder(dt_next(batoms))];
	 dt_addbond(at1, at2, dt_bondorder(mbond));
	 order = dt_bondorder(mbond);
	 dt_setimp_hcount(at1, dt_imp_hcount(at1)-order);  /* fix hydrogen counts */
	 dt_setimp_hcount(at2, dt_imp_hcount(at2)-order);
      }

      /* set atom attributes of transformed SMILES */
      dt_reset(matoms);
      for (matom = dt_next(matoms); matom != NULL_OB; matom = dt_next(matoms))
	 if (dt_number(matom)     != 0  ||
	     dt_imp_hcount(matom) != 0  ||
	     dt_charge(matom)     != 0)
	 {
	    oldnum = dt_number(mapping[dt_arborder(matom)]);
	    dt_setnumber(mapping[dt_arborder(matom)], dt_number(matom));
	    dt_setcharge(mapping[dt_arborder(matom)], dt_charge(matom));
	    dt_setweight(mapping[dt_arborder(matom)], dt_weight(matom));
	    if (dt_arborder(matom) > nkeep  ||  oldnum != dt_number(matom))
	       dt_setimp_hcount(mapping[dt_arborder(matom)], dt_imp_hcount(matom));
	 }

      /* get rid of streams */
      dt_dealloc(matoms); dt_dealloc(mbonds);

      if (flags & DROP_PIECES)
      {
	 if (debug) fprintf(stderr, "dropping pieces\n");
	 atoms = dt_stream(mol, TYP_ATOM);
	 for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	    dt_setarborder(atom, dt_count(mol, TYP_ATOM)+1);
	 
	 dt_setarborder(mapping[1], 0);

	 dt_mod_off(mol);
	 tmpsmi = dt_arbsmiles(&tmpsmilen, mol, TRUE);
	 strncpy(buffer, tmpsmi, tmpsmilen);
	 buffer[tmpsmilen] = '\0';
	 for (cp = buffer; *cp; cp++)
	    if ((*cp) == '.')
	    {
	       if (debug) fprintf(stderr, "cutting buffer\n");
	       (*cp) = '\0';
	    }
	 dt_dealloc(mol);
	 mol = StrToSmiles(buffer);
      }
      else
	 dt_mod_off(mol);

      free((void *) mapping);
   }	/* while ((nmatch < MAXMATCH  &&  flags & EXHAUSTIVE) || nmatch < 1) */

   if (nmatch == 0  &&  !(flags & EXHAUSTIVE))
   {
      dt_dealloc(mol);
      return (NULL_OB);
   }

   return (mol);
}

/*============================================================================
 *
 *  TfmTalk()
 *
 *  This is the main routine which handles the conversation with the
 *  client of this progob. It receives a handle to a sequence of strings
 *  which is the query sent by the client. It then computes the answer
 *  and packages it into another sequence of strings which is then returned.
 *
 *  The current version accepts sequences of strings but interprets only
 *  the first one.
 *
 *  The basic syntax of the commands is:
 *
 *  <OPCODE> ':' <SPACE> <PAR1> {<SPACE> <PARi>} <EOLN>
 *
 *  <SPACE> can be any sequence of blanks and tabs.
 *
 *  The following opcodes are implemented:
 *
 *  DEBUGON:	Enables debugging output to stderr.
 *
 *  DEBUGOFF:	Disables debugging output to stderr.
 *
 *  TRNCLEAR:	Clears the list of saved transforms. This command
 *		doesn't have
 *		parameters.
 *
 *  TRANSFORM:	This command saves a transformation of the form
 *
 *			<SMARTS> <SMILES> <nkeep> { <flag>}
 *
 *		which is meant to first match SMARTS and then replace the
 *		instance with SMILES deleting any atoms which are not part
 *		of the first nkeep in the match.
 *		
 *		The sequence of flags { <flag>} controls post processing
 *		after the transformation is applied. The following flags
 *		are currently defined:
 *
 *		DROP_PIECES	1 Causes all fragments of the structure
 *				  to be removed that are not connected to
 *				  the first atom of the SMARTS pattern.
 *
 *		EXHAUSTIVE	2 Apply the transformation until there is
 *				  no match. Only MAXMATCH cycles will be made.
 *				  Contrary to the non-exahustive mode,
 *				  the transformation will always return
 *				  a SMILES, except for errors in the
 *				  transformation itself.
 *
 *		REPLACE_TFMS	4 Do a TRNCLEAR before this transform is
 *				  added. This is meant to support one tfm
 *				  at a time processing.
 *
 *		UNIQUE_MATCH	8 The transform only matches if there is
 *				  either only one match of the target
 *				  SMARTS or if the resulting products
 *				  are identical. This is useful for
 *				  processing symmetrical reagents.
 *
 *		You can save up to MAXTRN transforms. They are kept in
 *		an array and the list is cleared with the command. TRNCLEAR:.
 *
 *  TRNSMILES:	Applies the saved TRANSFORMs to the SMILES given as
 *		its parameter. Transforms which have the EXHAUSTIVE flag
 *		set will be applied as is. If a transform is non-exhaustive,
 *		the SMILES is not modified if the transform didn't match.
 *
 *		The syntax of the line is
 *
 *			TRNSMILES: <SMILES> { <flag>}
 *
 *		The following flags are defined:
 *
 *		PREPEND_PRECURSOR Prepend the SMILES of the original molecule
 *				  to the resulting SMILES, separated by a '.'
 *				  character.
 *
 *		SMARTS_CHECK	  Perform a check against GOODSMARTS and
 *				  and BADSMARTS patterns before transformation
 *				  is attempted and POSTGOOD and POSTBAD
 *				  patterns after the transformation.
 *
 *		MDEF		  Reorder the SMILES such that [*-] comes
 *				  first, [*-2] comes last, and [*+n] is not
 *				  first. Then substitute [*+n] with &n and
 *				  remove [*-] and [*-2].
 *
 *  GOODSMARTS:	Has a white space separated list of SMARTS patterns as
 *		its parameters. This list is saved to be used by subsequent
 *		CHKSMILES commands or if the SMARTS_CHECK option is set
 *		for a TRNSMILES command.
 *
 *		A SMILES will pass the test if any of the listed
 *		SMARTS patterns matches.
 *
 *  BADSMARTS:	Has a white space separated list of SMARTS patterns as
 *		its parameters. This list will be used by the subsequent
 *		CHKSMILES commands of if the SMARTS_CHECK option is set
 *		for a TRNSMILES command.
 *
 *		A SMILES passes if none of the SMARTS patterns matches.
 *
 *  POSTGOOD:	Works like the GOODSMARTS: command except that the test
 *		is only applied for TRNSMILES commands after the
 *		transformation has been applied.
 *
 *  POSTBAD:	Works like the BADSMARTS: command except that the test
 *		is only applied for TRNSMILES commands after the
 *		transformation has been applied.
 *
 *  CHKSMILES:	Tests the SMILES string given as its parameter against
 *		the current GOODSMARTS and BADSMARTS lists. It returns
 *		the string "OK" if the SMILES passes the test and the
 *		empty sequence if not.
 */

struct tfm_desc_t		/* transform store */
   {
      dt_Handle  trnpat;
      dt_Handle  trnmol;
      dt_Integer nkeep;
      long       trnflags;
   } transforms[MAXTFM];
int ntfm = 0;

static dt_Handle TfmTalk(dt_Handle sosin)
{
   static long       trnflags = 0;	/* post processing flags */

   dt_Handle sob;			/* temporary string object */
   char *str;
   char str_buffer[2000];
   char *dy_str;

   dt_Integer lens;
   int n, i;

   dt_Handle sosout;			/* result object */
   dt_Handle mol, molout;		/* molecule object */
   char *retsmi;
   dt_Integer retsmilen;
   char buffer[1000];

   /* sequences of SMARTS pattern objects to be used by CHECKSMILES       */
   /* and before the transformation in TRNSMILES with SMARTS_CHECK option */
   static dt_Handle goodseq  = NULL_OB;
   static dt_Handle badseq   = NULL_OB;

   /* sequences of SMARTS pattern objects to be used by the post */
   /* transformation check in TRNSMILES with SMARTS_CHECK option */
   static dt_Handle postgood = NULL_OB;
   static dt_Handle postbad  = NULL_OB;

   dt_Handle atom, atoms;
   dt_Handle neighbour, neighbours;
   dt_Handle neighbourbond, neighbourbonds;
   int changed, minorder;

   int prepend_precursor;
   int smarts_checking;
   int convert_to_mdef;

   sosout = dt_alloc_seq();

   if (debug) fprintf(stderr,"\nsosin = %d\n", sosin);
   if (sosin == NULL_OB) return (sosout);

   sob = dt_next(sosin);			/* fetch first string. */
   if (debug) fprintf(stderr,"\nsob = %d\n", sob);
   if (sob == NULL_OB) return (sosout);

   dy_str = dt_stringvalue(&lens, sob); 	/* fetch command */
   if (lens > 2000) return (sosout);

   strncpy(str_buffer, dy_str, lens); 		/* copy Daylight buffer */
   str_buffer[lens] = '\0';			/* private string       */
   str = str_buffer;

   if (str != NULL) while (isspace(*str)) str++;

   if (debug) fprintf(stderr,"\nINPUT = '%.*s'\n", lens, str);

   if (strncmp(str, "DEBUGON:", strlen("DEBUGON:")) == 0)
   {
      fprintf(stderr, "Enabling debug output\n");
      debug = TRUE;
   }
   else if (strncmp(str, "DEBUGOFF:", strlen("DEBUGOFF:")) == 0)
   {
      debug = FALSE;
   }
   else if (strncmp(str, "TRNCLEAR:", strlen("TRNCLEAR:")) == 0)
   {					/* clear the transform store */

      for (i=0; i<ntfm; i++)
      {
	 if (transforms[i].trnpat != NULL_OB) dt_dealloc(transforms[i].trnpat);
	 if (transforms[i].trnmol != NULL_OB) dt_dealloc(transforms[i].trnmol);
	 transforms[i].trnpat == NULL_OB; transforms[i].trnmol == NULL_OB;
	 transforms[i].nkeep == 0; transforms[i].trnflags == 0;
      }
      ntfm = 0;
   }
   else if (strncmp(str, "TRANSFORM:", strlen("TRANSFORM:")) == 0)
   {					/* add a new transformation */
      if (ntfm >= MAXTFM) return (sosout);

      if (debug) fprintf(stderr, "decoding '%.*s'\n", lens, str);

      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL)
      {
	 fprintf(stderr, "error splitting off SMARTS\n");
	 return (sosout);
      }
      if (debug) fprintf(stderr, "SMARTS = %s\n", str);
      mol = dt_smartin(strlen(str), str);
      if (mol == NULL_OB)
      {
	 fprintf(stderr, "error parsing SMARTS '%s'\n", str);
	 return (sosout);
      }
      transforms[ntfm].trnpat = mol;

      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL)
      {
	 fprintf(stderr, "error splitting off result SMILES\n");
	 return (sosout);
      }
      mol = StrToSmiles(str);
      if (mol == NULL_OB)
      {
	 fprintf(stderr, "error parsing SMILES '%s'\n", str);
	 return (sosout);
      }
      transforms[ntfm].trnmol = mol;
      if (debug) fprintf(stderr, "mol = '%s'\n", DebugString(mol));

      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL)
      {
	 fprintf(stderr, "error splitting off nkeep\n");
	 return (sosout);
      }
      if (sscanf(str, "%d", &n) <= 0)
      {
	 fprintf(stderr, "error parsing nkeep = '%s'\n", str);
	 return (sosout);
      }
      transforms[ntfm].nkeep = n;

      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      transforms[ntfm].trnflags = 0;
      while (str != NULL)
      {
	 if (debug) fprintf(stderr, "residual = '%s'\n", str);
	 if (0 == strncmp(str, "DROP_PIECES", strlen("DROP_PIECES")))
	 {
	    if (debug) fprintf(stderr, "TRANSFORM: DROP_PIECES\n");
	    transforms[ntfm].trnflags |= DROP_PIECES;
	 }
	 else if (0 == strncmp(str, "EXHAUSTIVE", strlen("EXHAUSTIVE")))
	 {
	    if (debug) fprintf(stderr, "TRANSFORM: EXHAUSTIVE\n");
	    transforms[ntfm].trnflags |= EXHAUSTIVE;
	 }
	 else if (0 == strncmp(str, "REPLACE_TFMS", strlen("REPLACE_TFMS")))
	 {
	    if (debug) fprintf(stderr, "TRANSFORM: REPLACE_TFMS\n");
	    transforms[ntfm].trnflags |= REPLACE_TFMS;
	 }
	 else if (0 == strncmp(str, "UNIQUE_MATCH", strlen("UNIQUE_MATCH")))
	 {
	    if (debug) fprintf(stderr, "TRANSFORM: UNIQUE_MATCH\n");
	    transforms[ntfm].trnflags |= UNIQUE_MATCH;
	 }
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }

      if (transforms[ntfm].trnflags & REPLACE_TFMS)
      {
	 for (i=0; i<ntfm; i++)
	 {
	    if (transforms[i].trnpat != NULL_OB)
	       dt_dealloc(transforms[i].trnpat);
	    if (transforms[i].trnmol != NULL_OB)
	       dt_dealloc(transforms[i].trnmol);
	    transforms[i].trnpat == NULL_OB; transforms[i].trnmol == NULL_OB;
	    transforms[i].nkeep == 0; transforms[i].trnflags == 0;
	 }
	 transforms[0].trnpat = transforms[ntfm].trnpat;
	 transforms[0].trnmol = transforms[ntfm].trnmol;
	 transforms[0].nkeep  = transforms[ntfm].nkeep ;
	 transforms[0].trnflags = transforms[ntfm].trnflags;
	 ntfm = 1;
      }
      else
	 ntfm++;
   }
   else if (strncmp(str, "TRNSMILES:", strlen("TRNSMILES:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      mol = StrToSmiles(str);

      if (mol == NULL_OB)
      {
	 fprintf(stderr, "TRNSMILES: problems converting SMILES");
	 return (sosout);
      }

      if (debug)
	 fprintf(stderr, "TRNSMILES: input-SMILES = '%s' -> '%s'\n",
		 str, DebugString(mol));

      prepend_precursor = FALSE;		/* read flags */
      smarts_checking = FALSE;
      convert_to_mdef = FALSE;
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      while (str != NULL)
      {
	 if (0 == strncmp(str,
			  "PREPEND_PRECURSOR",
			  strlen("PREPEND_PRECURSOR")))
	    prepend_precursor = TRUE;
	 else if (0 == strncmp(str, "SMARTS_CHECK", strlen("SMARTS_CHECK")))
	    smarts_checking = TRUE;
	 else if (0 == strncmp(str, "MDEF", strlen("MDEF")))
	    convert_to_mdef = TRUE;
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }

      if (prepend_precursor)			/* show precursor */
      {
	 if (debug) fprintf(stderr, "prepending precursor\n");
	 retsmi = dt_arbsmiles(&retsmilen, mol, TRUE);
	 strncpy(buffer, retsmi, retsmilen); buffer[retsmilen] = '\0';
	 strcat(buffer, ".");
      }
      else
	 buffer[0] = '\0';

      if (smarts_checking  &&
	  (goodseq != NULL_OB  ||  badseq != NULL_OB))
      {
	 if (debug) fprintf(stderr, "Pretransform SMARTS checking\n");
	 if (!CheckSmiles(mol, goodseq, badseq))
	 {
	    dt_dealloc(mol);
	    return (sosout);
	 }
      }

      for (i=0; i<ntfm; i++)	/* apply the transformation as defined */
      {
	 molout = Transform(mol,
			    transforms[i].trnpat,
			    transforms[i].trnmol,
			    transforms[i].nkeep,
			    transforms[i].trnflags);
	 if (debug)
	 {
	    fprintf(stderr, "TRNSMILES: mol = %s\n", DebugString(mol));
	    fprintf(stderr, "TRNSMILES: trnpat = %d\n", transforms[i].trnpat);
	    fprintf(stderr, "TRNSMILES: trnmol = %s\n",
		    DebugString(transforms[i].trnmol));
	    fprintf(stderr, "TRNSMILES: nkeep = %d, trnflags = %d\n",
		    transforms[i].nkeep, transforms[i].trnflags);
	    fprintf(stderr, "TRNSMILES: molout = %s\n", DebugString(molout));
	 }

	 if (molout != NULL_OB)
	 {
	    dt_dealloc(mol); mol = molout; molout = NULL_OB;
	 }
      }

      if (smarts_checking)
      {
	 if (debug)
	    fprintf(stderr,
		    "Post SMARTS: good = (%d), bad = (%d)\n",
		    postgood, postbad);
	 if (postgood != NULL_OB  ||  postbad != NULL_OB)
	 {
	    if (debug) fprintf(stderr, "Posttransform SMARTS checking\n");
	    if (!CheckSmiles(mol, postgood, postbad))
	    {
	       dt_dealloc(mol);
	       dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	       return (sosout);
	    }
	 }
      }

      if (convert_to_mdef)
      {
	 if (debug) fprintf(stderr, "converting to MDEF\n");
	 /* First handle [*-2 case. 					*/
	 /* Algorithm:  1: Set all arborders to 0. 			*/
	 /*		2: Set arborder of [*-2 atom to high value. 	*/
	 /*		3: loop until there is no change 		*/
	 /*		  3a: find atom with non-zero arborder (ao). 	*/
	 /*		  3b: loop through its neighbours. 		*/
	 /*		    3ba: if arborder of neighbour == 0 then	*/
	 /*			   set to ao-1.				*/

	 atoms = dt_stream(mol, TYP_ATOM);
	 for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	    if (dt_number(atom) == 0  &&  dt_charge(atom) == (-2))
	       dt_setarborder(atom, dt_count(mol, TYP_ATOM)+2);
	    else
	       dt_setarborder(atom, 0);

	 do
	 {
	    changed = FALSE;
	    dt_reset(atoms);
	    for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	       if (dt_arborder(atom) != 0)	/* candidate */
	       {
		  neighbourbonds = dt_stream(atom, TYP_BOND);
		  for (neighbourbond = dt_next(neighbourbonds);
		       neighbourbond != NULL_OB;
		       neighbourbond = dt_next(neighbourbonds))
		  {
		     neighbours =
			dt_stream(neighbourbond, TYP_ATOM); /* !! TRICK !! */
		     for (neighbour = dt_next(neighbours);
			  neighbour != NULL_OB;
			  neighbour = dt_next(neighbours))
			if (dt_arborder(neighbour) == 0)
			{
			   changed = TRUE;
			   dt_setarborder(neighbour, dt_arborder(atom)-1);
			}
		  }
	       }
	 } while (changed);

	 /* Now find minimal arborder to place other [*??] atoms */
	 minorder = dt_count(mol, TYP_ATOM)+2;
	 dt_reset(atoms);
	 for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	    if (minorder > dt_arborder(atom)) minorder = dt_arborder(atom);
	 
	 /* Set arborder of [*-] atom(s?) to less than minorder - 1 */
	 dt_reset(atoms);
	 for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	    if (dt_number(atom) == 0  &&  dt_charge(atom) == (-1)) /* [*-] */
	       dt_setarborder(atom, minorder-1);

	 /* Set arborders of [*+n] atoms to one more than minorder to make */
	 /* sure that they don't come at the beginning. */
	 dt_reset(atoms);
	 for (atom = dt_next(atoms); atom != NULL_OB; atom = dt_next(atoms))
	    if (dt_number(atom) == 0  &&  dt_charge(atom) > 0)	/* [*+n] */
	       dt_setarborder(atom, minorder+1);
      }

      retsmi = dt_arbsmiles(&retsmilen, mol, TRUE);
      strcat(buffer, retsmi);

      if (convert_to_mdef)	/* do string substitution */
      {
	 if (strstr(buffer,"[*-]") != buffer  &&
	     strstr(buffer,"[*-]") != NULL)
	 {
	    fprintf(stderr,
		    "'From' attachment not at beginning of '%s'\n",
		    buffer);
	    dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	    return (sosout);
	 }
	 while (sub("[*-]", "", buffer))
	    if (strstr(buffer,"[*-]") != buffer  &&
		strstr(buffer,"[*-]") != NULL)
	    {
	       fprintf(stderr,
		       "'From' attachment not at beginning of '%s'\n",
		       buffer);
	       dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	       return (sosout);
	    }

	 if (strstr(buffer,"[*-2]") != NULL  &&
	     strstr(buffer,"[*-2]") != buffer+strlen(buffer)-strlen("[*-2]"))
	 {
	    fprintf(stderr,"'To' attachment not at end of '%s'\n", buffer);
	    dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	    return (sosout);
	 }
	 while (sub("[*-2]", "", buffer))
	    ;

	 if (strstr(buffer,"[*+") != NULL  &&  strstr(buffer,"[*+") == buffer)
	 {
	    fprintf(stderr,
		    "External connection at left end of SMILES '%s'\n",
		    buffer);
	    dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	    return (sosout);
	 }
	 if (strstr(buffer,".[*+") != NULL)
	 {
	    fprintf(stderr,
		    "External connection at start of fragment in SMILES '%s'\n",
		    buffer);
	    dt_dealloc(sosout); sosout = dt_alloc_seq();	/* clear result */
	    return (sosout);
	 }

	 /* fix isolated hydrogen problem ([*H+] => [H]&1) */
	 while (sub("[*H+]", "[H]&1", buffer))
	    ;

	 /* fix R-groups in parenthesis */
	 while (sub("([*+])", "&1", buffer))
	    ;
	 while (sub("([*+2])", "&2", buffer))
	    ;
	 while (sub("([*+3])", "&3", buffer))
	    ;
	 while (sub("([*+4])", "&4", buffer))
	    ;
	 while (sub("([*+5])", "&5", buffer))
	    ;
	 while (sub("([*+6])", "&6", buffer))
	    ;
	 while (sub("([*+7])", "&7", buffer))
	    ;
	 while (sub("([*+8])", "&8", buffer))
	    ;
	 while (sub("([*+9])", "&9", buffer))
	    ;

	 /* fix ordinary R-groups */
	 while (sub("[*+]", "&1", buffer))
	    ;
	 while (sub("[*+2]", "&2", buffer))
	    ;
	 while (sub("[*+3]", "&3", buffer))
	    ;
	 while (sub("[*+4]", "&4", buffer))
	    ;
	 while (sub("[*+5]", "&5", buffer))
	    ;
	 while (sub("[*+6]", "&6", buffer))
	    ;
	 while (sub("[*+7]", "&7", buffer))
	    ;
	 while (sub("[*+8]", "&8", buffer))
	    ;
	 while (sub("[*+9]", "&9", buffer))
	    ;

      }

      dt_append(sosout, dt_alloc_string(strlen(buffer), buffer));

      dt_dealloc(mol);
   }
   else if (strncmp(str, "GOODSMARTS:", strlen("GOODSMARTS:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      if (goodseq) dt_dealloc(goodseq);

      goodseq = dt_alloc_seq();
      while (str != NULL)
      {
	 mol = dt_smartin(strlen(str), str);
	 if (debug) fprintf(stderr, "GOODSMARTS: %s (%d)\n", str, mol);
	 if (mol != NULL_OB) dt_append(goodseq, mol);
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }
   }
   else if (strncmp(str, "BADSMARTS:", strlen("BADSMARTS:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      if (badseq) dt_dealloc(badseq);

      badseq = dt_alloc_seq();
      while (str != NULL)
      {
	 mol = dt_smartin(strlen(str), str);
	 if (debug) fprintf(stderr, "BADSMARTS: %s (%d)\n", str, mol);
	 if (mol != NULL_OB) dt_append(badseq, mol);
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }
   }
   else if (strncmp(str, "POSTGOOD:", strlen("POSTGOOD:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      if (postgood) dt_dealloc(postgood);

      postgood = dt_alloc_seq();
      while (str != NULL)
      {
	 mol = dt_smartin(strlen(str), str);
	 if (debug) fprintf(stderr, "POSTGOOD: %s (%d)\n", str, mol);
	 if (mol != NULL_OB) dt_append(postgood, mol);
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }
   }
   else if (strncmp(str, "POSTBAD:", strlen("POSTBAD:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      if (postbad) dt_dealloc(postbad);

      postbad = dt_alloc_seq();
      while (str != NULL)
      {
	 mol = dt_smartin(strlen(str), str);
	 if (debug) fprintf(stderr, "POSTBAD: %s (%d)\n", str, mol);
	 if (mol != NULL_OB) dt_append(postbad, mol);
	 str = strpbrk(str, " \t");
	 if (str != NULL) while (isspace(*str)) str++;
      }
   }
   else if (strncmp(str, "CHKSMILES:", strlen("CHKSMILES:")) == 0)
   {
      str = strpbrk(str, " \t"); if (str != NULL) while (isspace(*str)) str++;
      if (str == NULL) return (sosout);
      mol = StrToSmiles(str);
      if (mol == NULL_OB) return (sosout);
      if (CheckSmiles(mol, goodseq, badseq))
      {
	 dt_append(sosout, dt_alloc_string(strlen("OK"), "OK"));
      }
   }

   return (sosout);
}

/*============================================================================
 *  main() -- just pass control to event loop with no welcome message
 */

main()
{
   du_pipe_loop(NULL_OB, TfmTalk);
}
